#include <gtk/gtk.h>
#include <stdlib.h>
#include <stdio.h>
#include "defines.h"
#include "vpdefines.h"
#include "globals.h"
#include "globals_gtk.h"

#include "protos_volume.h"
#include "protos_gui.h"
#include "protos_menus.h"


void create_menu(GtkWidget *window, GtkWidget **menubar){
   GtkItemFactory *item_factory;
   GtkAccelGroup *accel_group;
   static GtkItemFactoryEntry menu_items[] = {
      { "/_File",             NULL,         NULL, 0, "<Branch>" },
      { "/File/_Open",        "<control>O", create_file_window, 0, NULL },
      { "/File/_Save Image",  "<control>S", create_save_window, 0, NULL },
      { "/File/_Load Config", "<control>L", create_lconfig_window, 0, NULL},
      { "/File/Save Config",  NULL,         create_sconfig_window, 0, NULL},
      { "/File/sep1",         NULL,         NULL, 0, "<Separator>" },
      { "/File/_Quit",         "<control>Q", gtk_main_quit, 0, NULL },
      { "/_Options",          NULL,         NULL, 0, "<Branch>" },
      { "/Options/View",      NULL,         create_view_window, 0, NULL },
      { "/_Tools",            NULL,         NULL, 0, "<Branch>" },
      { "/Tools/Histogram",   "<control>H", create_histogram, 0, NULL },
      { "/_Help",             NULL,         NULL, 0, "<LastBranch>" },
      { "/_Help/About",       NULL,         NULL, 0, NULL },
   };
   gint nmenu_items = sizeof (menu_items) / sizeof (menu_items[0]);


   accel_group = gtk_accel_group_new ();

   item_factory = gtk_item_factory_new (GTK_TYPE_MENU_BAR, "<main>", accel_group);
   gtk_item_factory_create_items (item_factory, nmenu_items, menu_items, NULL);
   gtk_accel_group_attach (accel_group, GTK_OBJECT (window));
   *menubar = gtk_item_factory_get_widget (item_factory, "<main>");
}

void open_ok_sel (GtkWidget *w, GtkFileSelection *fs){
   if(_rawdata != NULL) free(_rawdata);
   if(vpc != NULL) vpDestroyContext(vpc);
   if(shade_table != NULL) free(shade_table);
   if(weight_table != NULL) free(weight_table);
   if(image != NULL) free(image);

   _rawdata = NULL;
   vpc = NULL;
   shade_table = NULL;
   weight_table = NULL;
   image = NULL;

   viewVolume[0][0] = -0.5;
   viewVolume[0][1] = 0.5;
   viewVolume[1][0] = -0.5;
   viewVolume[1][1] = 0.5;
   viewVolume[2][0] = -0.5;
   viewVolume[2][1] = 0.5;

   imageScaleX = 1.0;
   imageScaleY = 1.0;

   if(readVolume(gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs)) ) != 1){
      gtk_widget_destroy(GTK_WIDGET(fs));
      return;
   }
   load_volume();
   gtk_widget_destroy(GTK_WIDGET(fs));
}

void save_ok_sel (GtkWidget *w, GtkFileSelection *fs){
   if(image == NULL){
      fprintf(stderr, "Image buffer is empty\n");
      return;
   }

   StorePPM(image, _VolumeDimensions.x *imageScaleX, _VolumeDimensions.y * imageScaleY, 
            gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs)));
   gtk_widget_destroy(GTK_WIDGET(fs));
}

void loadConfig_ok_sel(GtkWidget *w, GtkFileSelection *fs){
   FILE *in;
   int i;
   char string[20];
   GtkAdjustment *adj;

   /* read in the configuration file */
   in = fopen(gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs)), "rb");
   fread(&numDensity, sizeof(int), 1, in);
   fread(&DensityRampX, sizeof(int), 18, in);
   fread(&DensityRampY, sizeof(float), 18, in);
   fread(&numGradient, sizeof(int), 1, in);
   fread(&GradientRampX, sizeof(int), 7, in);
   fread(&GradientRampY, sizeof(float), 7, in);

   for(i = 0; i < 6; i++){
      fread(&matAmb[i], sizeof(C_RGBtuple), 1, in);
      fread(&matDiff[i], sizeof(C_RGBtuple), 1, in);
      fread(&matSpec[i], sizeof(C_RGBtuple), 1, in);
      fread(&matShin[i], sizeof(C_RGBtuple), 1, in);
   }

   for(i = 0; i < 6; i++){
      fread(&lightColor[i], sizeof(C_RGBtuple), 1, in);
      fread(&lightDir[i], sizeof(C_FloatTuple), 1, in);
   }

   fread(&numMaterials, sizeof(int), 1, in);
   fread(&numLights, sizeof(int), 1, in);
   fread(&weightTable, sizeof(int), 12, in);
   fread(&viewVolume, sizeof(float), 6, in);
   fread(&imageScaleX, sizeof(float), 1, in);
   fread(&imageScaleY, sizeof(float), 1, in);

   fclose(in);
   gtk_widget_destroy(GTK_WIDGET(fs));
   
   /* update the gui */
   sprintf(string, "%d", numDensity);
   gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(numDensityCombo)->entry), string);
   sprintf(string, "%d", numGradient);
   gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(numGradientCombo)->entry), string);
   sprintf(string, "%d", numMaterials);
   gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(numMaterialsCombo)->entry), string);
   sprintf(string, "%d", numLights);
   gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(numLightsCombo)->entry), string);
   
   for(i = 0; i < numDensity/2; i++){
      sprintf(string, "%d", DensityRampX[i*2]);
      gtk_entry_set_text(GTK_ENTRY(dpx[i][0]), string);
      sprintf(string, "%d", DensityRampX[i*2+1]);
      gtk_entry_set_text(GTK_ENTRY(dpx[i][1]), string);
   }

   for(i = 0; i < numDensity/2; i++){
      sprintf(string, "%f", DensityRampY[i*2]);
      gtk_entry_set_text(GTK_ENTRY(dpy[i][0]), string);
      sprintf(string, "%f", DensityRampY[i*2+1]);
      gtk_entry_set_text(GTK_ENTRY(dpy[i][1]), string);
   }

   for(i = 0; i < numGradient; i++){
      sprintf(string, "%d", GradientRampX[i]);
      gtk_entry_set_text(GTK_ENTRY(gpx[i]), string);
   }

   for(i = 0; i < numGradient; i++){
      sprintf(string, "%f", GradientRampY[i]);
      gtk_entry_set_text(GTK_ENTRY(gpy[i]), string);
   }

   for(i = 0; i < numMaterials; i++){
      adj = gtk_range_get_adjustment(GTK_RANGE(matAmbScale[i][0]));
      gtk_adjustment_set_value(adj, matAmb[i].red);
      adj = gtk_range_get_adjustment(GTK_RANGE(matAmbScale[i][1]));
      gtk_adjustment_set_value(adj, matAmb[i].green);
      adj = gtk_range_get_adjustment(GTK_RANGE(matAmbScale[i][2]));
      gtk_adjustment_set_value(adj, matAmb[i].blue);
   
      adj = gtk_range_get_adjustment(GTK_RANGE(matDiffScale[i][0]));
      gtk_adjustment_set_value(adj, matDiff[i].red);
      adj = gtk_range_get_adjustment(GTK_RANGE(matDiffScale[i][1]));
      gtk_adjustment_set_value(adj, matDiff[i].green);
      adj = gtk_range_get_adjustment(GTK_RANGE(matDiffScale[i][2]));
      gtk_adjustment_set_value(adj, matDiff[i].blue);
   
      adj = gtk_range_get_adjustment(GTK_RANGE(matSpecScale[i][0]));
      gtk_adjustment_set_value(adj, matSpec[i].red);
      adj = gtk_range_get_adjustment(GTK_RANGE(matSpecScale[i][1]));
      gtk_adjustment_set_value(adj, matSpec[i].green);
      adj = gtk_range_get_adjustment(GTK_RANGE(matSpecScale[i][2]));
      gtk_adjustment_set_value(adj, matSpec[i].blue);
   
      adj = gtk_range_get_adjustment(GTK_RANGE(matShinyScale[i]));
      gtk_adjustment_set_value(adj, matShin[i].red);
   }

   for(i = 0; i < numLights; i++){
      adj = gtk_range_get_adjustment(GTK_RANGE(lightDirScale[i][0]));
      gtk_adjustment_set_value(adj, lightDir[i].x);
      adj = gtk_range_get_adjustment(GTK_RANGE(lightDirScale[i][1]));
      gtk_adjustment_set_value(adj, lightDir[i].y);
      adj = gtk_range_get_adjustment(GTK_RANGE(lightDirScale[i][2]));
      gtk_adjustment_set_value(adj, lightDir[i].z);

      adj = gtk_range_get_adjustment(GTK_RANGE(lightColorScale[i][0]));
      gtk_adjustment_set_value(adj, lightColor[i].red);
      adj = gtk_range_get_adjustment(GTK_RANGE(lightColorScale[i][1]));
      gtk_adjustment_set_value(adj, lightColor[i].green);
      adj = gtk_range_get_adjustment(GTK_RANGE(lightColorScale[i][2]));
      gtk_adjustment_set_value(adj, lightColor[i].blue);
   }

   for(i = 0; i < numMaterials; i++){
      sprintf(string, "%d", weightTable[i][0]);
      gtk_entry_set_text(GTK_ENTRY(weightEntry[i][0]), string);
      sprintf(string, "%d", weightTable[i][1]);
      gtk_entry_set_text(GTK_ENTRY(weightEntry[i][1]), string);
   }  



}

void saveConfig_ok_sel(GtkWidget *w, GtkFileSelection *fs){
   FILE *out;
   int i;

   out = fopen(gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs)), "wb");
   fwrite(&numDensity, sizeof(int), 1, out);
   fwrite(&DensityRampX, sizeof(int), 18, out);
   fwrite(&DensityRampY, sizeof(float), 18, out);
   fwrite(&numGradient, sizeof(int), 1, out);
   fwrite(&GradientRampX, sizeof(int), 7, out);
   fwrite(&GradientRampY, sizeof(float), 7, out);

   for(i = 0; i < 6; i++){
      fwrite(&matAmb[i], sizeof(C_RGBtuple), 1, out);
      fwrite(&matDiff[i], sizeof(C_RGBtuple), 1, out);
      fwrite(&matSpec[i], sizeof(C_RGBtuple), 1, out);
      fwrite(&matShin[i], sizeof(C_RGBtuple), 1, out);
   }
   
   for(i = 0; i < 6; i++){
      fwrite(&lightColor[i], sizeof(C_RGBtuple), 1, out);
      fwrite(&lightDir[i], sizeof(C_FloatTuple), 1, out);
   }

   fwrite(&numMaterials, sizeof(int), 1, out);
   fwrite(&numLights, sizeof(int), 1, out);
   fwrite(&weightTable, sizeof(int), 12, out);
   fwrite(&viewVolume, sizeof(float), 6, out);
   fwrite(&imageScaleX, sizeof(float), 1, out);
   fwrite(&imageScaleY, sizeof(float), 1, out);
   
   fclose(out);
   gtk_widget_destroy(GTK_WIDGET(fs));
}


void create_file_window(){
   GtkWidget *filew;

   filew = gtk_file_selection_new("Open File");
   gtk_signal_connect(GTK_OBJECT (GTK_FILE_SELECTION (filew)->ok_button), 
                      "clicked", (GtkSignalFunc) open_ok_sel, filew);
   gtk_signal_connect_object(GTK_OBJECT (GTK_FILE_SELECTION(filew)->cancel_button),
                             "clicked", (GtkSignalFunc) gtk_widget_destroy, GTK_OBJECT (filew));
   gtk_widget_show(filew);

}

void create_save_window(){
   GtkWidget *filew;

   if(image == NULL) return;

   filew = gtk_file_selection_new("Save Image");
   gtk_signal_connect(GTK_OBJECT (GTK_FILE_SELECTION (filew)->ok_button), 
                      "clicked", (GtkSignalFunc) save_ok_sel, filew);
   gtk_signal_connect_object(GTK_OBJECT (GTK_FILE_SELECTION(filew)->cancel_button),
                             "clicked", (GtkSignalFunc) gtk_widget_destroy, GTK_OBJECT (filew));
   gtk_widget_show(filew);
}

void create_lconfig_window(){
   GtkWidget *filew;

   filew = gtk_file_selection_new("Load Config");
   gtk_signal_connect(GTK_OBJECT (GTK_FILE_SELECTION (filew)->ok_button), 
                      "clicked", (GtkSignalFunc) loadConfig_ok_sel, filew);
   gtk_signal_connect_object(GTK_OBJECT (GTK_FILE_SELECTION(filew)->cancel_button),
                             "clicked", (GtkSignalFunc) gtk_widget_destroy, GTK_OBJECT (filew));
   gtk_widget_show(filew);
}

void create_sconfig_window(){
   GtkWidget *filew;

   filew = gtk_file_selection_new("Save Config");
   gtk_signal_connect(GTK_OBJECT (GTK_FILE_SELECTION (filew)->ok_button), 
                      "clicked", (GtkSignalFunc) saveConfig_ok_sel, filew);
   gtk_signal_connect_object(GTK_OBJECT (GTK_FILE_SELECTION(filew)->cancel_button),
                             "clicked", (GtkSignalFunc) gtk_widget_destroy, GTK_OBJECT (filew));
   gtk_widget_show(filew);
}

void create_view_window(){
   GtkWidget *window, *hbox, *vbox, *label, *button;

   window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
   gtk_window_set_title(GTK_WINDOW(window), "View Options");
   gtk_container_border_width(GTK_CONTAINER(window), 2);
   
   vbox = gtk_vbox_new(FALSE, 10);
   
   /* Create label */
   hbox = gtk_hbox_new(FALSE, 5);
   label = gtk_label_new("View Volume Coordinates");
   gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, FALSE, 0);
   gtk_widget_show(label);

   gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, FALSE, 0);
   gtk_widget_show(hbox);

   /* Create first row of view volume entry boxes */
   hbox = gtk_hbox_new(FALSE, 5);
   label = gtk_label_new("X");
   vvEntry[0][0] = gtk_entry_new();
   gtk_entry_set_max_length(GTK_ENTRY(vvEntry[0][0]), 5);
   gtk_widget_set_usize(vvEntry[0][0], 40, 22);
   vvEntry[0][1] = gtk_entry_new();
   gtk_entry_set_max_length(GTK_ENTRY(vvEntry[0][1]), 5);
   gtk_widget_set_usize(vvEntry[0][1], 40, 22);

   gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, FALSE, 0);
   gtk_box_pack_start(GTK_BOX(hbox), vvEntry[0][0], TRUE, FALSE, 0);
   gtk_box_pack_start(GTK_BOX(hbox), vvEntry[0][1], TRUE, FALSE, 0);
   gtk_widget_show(label);
   gtk_widget_show(vvEntry[0][0]);
   gtk_widget_show(vvEntry[0][1]);

   gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, FALSE, 0);
   gtk_widget_show(hbox);
   
   /* Create second row of view volume entry boxes */
   hbox = gtk_hbox_new(FALSE, 5);
   label = gtk_label_new("Y");
   vvEntry[1][0] = gtk_entry_new();
   gtk_entry_set_max_length(GTK_ENTRY(vvEntry[1][0]), 5);
   gtk_widget_set_usize(vvEntry[1][0], 40, 22);
   vvEntry[1][1] = gtk_entry_new();
   gtk_entry_set_max_length(GTK_ENTRY(vvEntry[1][1]), 5);
   gtk_widget_set_usize(vvEntry[1][1], 40, 22);

   gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, FALSE, 0);
   gtk_box_pack_start(GTK_BOX(hbox), vvEntry[1][0], TRUE, FALSE, 0);
   gtk_box_pack_start(GTK_BOX(hbox), vvEntry[1][1], TRUE, FALSE, 0);
   gtk_widget_show(label);                   
   gtk_widget_show(vvEntry[1][0]);
   gtk_widget_show(vvEntry[1][1]);

   gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, FALSE, 0);
   gtk_widget_show(hbox);
   
   /* Create third row of view volume entry boxes */
   hbox = gtk_hbox_new(FALSE, 5);
   label = gtk_label_new("Z");
   vvEntry[2][0] = gtk_entry_new();
   gtk_entry_set_max_length(GTK_ENTRY(vvEntry[2][0]), 5);
   gtk_widget_set_usize(vvEntry[2][0], 40, 22);
   vvEntry[2][1] = gtk_entry_new();
   gtk_entry_set_max_length(GTK_ENTRY(vvEntry[2][1]), 5);
   gtk_widget_set_usize(vvEntry[2][1], 40, 22);

   gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, FALSE, 0);
   gtk_box_pack_start(GTK_BOX(hbox), vvEntry[2][0], TRUE, FALSE, 0);
   gtk_box_pack_start(GTK_BOX(hbox), vvEntry[2][1], TRUE, FALSE, 0);
   gtk_widget_show(label);                   
   gtk_widget_show(vvEntry[2][0]);
   gtk_widget_show(vvEntry[2][1]);

   gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, FALSE, 0);
   gtk_widget_show(hbox);
   
   /* Create apply button */
   hbox = gtk_hbox_new(FALSE, 5);
   button = gtk_button_new_with_label("Apply");
   gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(apply_view_window), NULL);
   gtk_signal_connect_object(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(window));
   gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, FALSE, 0);
   gtk_widget_show(button);

   button = gtk_button_new_with_label("Cancel");
   gtk_signal_connect_object(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(window));
   gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, FALSE, 0);
   gtk_widget_show(button);

   gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, FALSE, 0);
   gtk_widget_show(hbox);
   
   /* Add it all to the window */
   gtk_container_add(GTK_CONTAINER(window), GTK_WIDGET(vbox));
   gtk_widget_show(GTK_WIDGET(vbox));
   gtk_widget_show(GTK_WIDGET(window));
}

void apply_view_window(GtkWidget *button){
   
   viewVolume[0][0] = atof( (char*) gtk_editable_get_chars(GTK_EDITABLE(vvEntry[0][0]), 0, -1));
   viewVolume[0][1] = atof( (char*) gtk_editable_get_chars(GTK_EDITABLE(vvEntry[0][1]), 0, -1));
   viewVolume[1][0] = atof( (char*) gtk_editable_get_chars(GTK_EDITABLE(vvEntry[1][0]), 0, -1));
   viewVolume[1][1] = atof( (char*) gtk_editable_get_chars(GTK_EDITABLE(vvEntry[1][1]), 0, -1));
   viewVolume[2][0] = atof( (char*) gtk_editable_get_chars(GTK_EDITABLE(vvEntry[2][0]), 0, -1));
   viewVolume[2][1] = atof( (char*) gtk_editable_get_chars(GTK_EDITABLE(vvEntry[2][1]), 0, -1));
   
   vpCurrentMatrix(vpc, VP_PROJECT);
   vpIdentityMatrix(vpc);
   vpWindow(vpc, VP_PARALLEL, viewVolume[0][0], viewVolume[0][1], viewVolume[1][0], viewVolume[1][1], 
            viewVolume[2][0], viewVolume[2][1]);
   vpCurrentMatrix(vpc, VP_MODEL);

   imageScaleX = (viewVolume[0][1] - viewVolume[0][0]) / 1.0;
   imageScaleY = (viewVolume[1][1] - viewVolume[1][0]) / 1.0;
   free(image);
   image = malloc(_VolumeDimensions.x * imageScaleX * _VolumeDimensions.y * imageScaleY * 3);
   fprintf(stderr, "%.2f - %.2f\n", imageScaleX, imageScaleY);

   
}


void histogram_expose(GtkWidget *widget, GdkEventExpose *event, gpointer user_data){
   int i, max, scale;
   
   gdk_draw_rectangle(widget->window, widget->style->black_gc, TRUE, 0, 0,
                      widget->allocation.width, widget->allocation.height);
   
   max = 0;
   for(i = 0; i < 256; i++){
      if(data[i] > max) max = data[i];
   }

   scale = max / widget->allocation.height;
   fprintf(stderr, "max = %d\n", max);
   fprintf(stderr, "scale = %d\n", scale);
   
   for(i = 0; i < 256; i++){
      gdk_draw_line(widget->window, widget->style->white_gc, i, widget->allocation.height, i,
                    widget->allocation.height - (data[i] / scale));
   }
}

void create_histogram(){
   GtkWidget *window, *drawing_area;
   unsigned char *cursor;
   int i, size;
   
   if(_rawdata == NULL) {
      fprintf(stderr, "rawdata is null\n");
      return;
   }

   /* ---- Gather the Data ---- */
   for(i = 0; i < 256; i++) data[i] = 0;
   cursor = _rawdata;
   size = _VolumeDimensions.x * _VolumeDimensions.y * _VolumeDimensions.z;
   for(i = 0; i < size; i++){
      data[*cursor]++;
      cursor++;
   }
   
   fprintf(stderr, "size = %d\n", size);
   for(i = 0; i < 256; i++) fprintf(stderr, "density %d = %d\n", i, data[i]);

   /* ---- Create the GUI ---- */
   window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
   gtk_window_set_title(GTK_WINDOW(window), "Histogram");
   gtk_container_border_width(GTK_CONTAINER(window), 2);
   /*gtk_signal_connect(GTK_OBJECT(window), "delete_event", GTK_SIGNAL_FUNC(), NULL)*/

   drawing_area = gtk_drawing_area_new();
   gtk_drawing_area_size(GTK_DRAWING_AREA(drawing_area), 256, 256);
   gtk_signal_connect (GTK_OBJECT (drawing_area), "expose-event", GTK_SIGNAL_FUNC (histogram_expose), NULL);
   
   gtk_container_add(GTK_CONTAINER(window), drawing_area);
   gtk_widget_show(drawing_area);

   gtk_widget_show(window);
}


void StorePPM(char *image, int width, int height, char *filename){

   FILE *image_fp;	/* file descriptor for image (output) */

   #define PPM_MAGIC1	'P'
   #define RPPM_MAGIC2	'6'

    if ((image_fp = fopen(filename, "w")) == NULL) {
	fprintf(stderr, "cannot open output file %s\n", filename);
	exit(1);
    }
    fprintf(image_fp, "%c%c\n%d %d\n%d\n", PPM_MAGIC1, RPPM_MAGIC2,
	    width, height, 255);
    fwrite(image, 1, width*height*3, image_fp);
    fclose(image_fp);
}

